/// Programa: Calibracion
/// Descripcion: Realiza la calibracion de dos camaras para su utilizacion
/// en un sistema de vision estereoscopica. La calibracion se realiza
/// mediante un patron, un tablero de ajedrez, el cual es imprescindible
/// para poder calibrar las camaras.
/// Autor: Carlos Castedo Hernandez
/// Fecha: 12/03/2017

#include <iostream>
#include <iomanip>
#include <stdio.h>
#include <opencv2/opencv.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/calib3d/calib3d.hpp>
#include "../include/config.h"

using namespace cv;
using namespace std;

int main()
{
    /// Calibracion
    vector<Mat> izq_imagenes, dcha_imagenes;            // Vectores de Mats
    Mat combo(FRAME_HEIGHT, 2 * FRAME_WIDTH, CV_8UC3);  // Imagen para visualizacin

    // Comprobacion imagenes residuales
    string filepath_izda = "../Data/Imagenes/Izquierda1.jpg";
    string filepath_dcha = "../Data/Imagenes/Derecha1.jpg";

    Mat dch = imread(filepath_dcha.c_str(),IMREAD_COLOR);
    Mat izq = imread(filepath_izda.c_str(),IMREAD_COLOR);
    if(!izq.empty() || !dch.empty()){
        cout << "Borre las imagenes de la carpeta Data/Imagenes" << endl;
        return 0;
    }

    VideoCapture capIzda(1), capDcha(0);    // Declaracin cmaras

    // Set tamao imagenes capturadas
    capIzda.set(CV_CAP_PROP_FRAME_HEIGHT, FRAME_HEIGHT);
    capIzda.set(CV_CAP_PROP_FRAME_WIDTH, FRAME_WIDTH);
    capDcha.set(CV_CAP_PROP_FRAME_HEIGHT, FRAME_HEIGHT);
    capDcha.set(CV_CAP_PROP_FRAME_WIDTH, FRAME_WIDTH);

    // Comprobacin imagen de cmaras
    if(!capIzda.open(1) || !capDcha.open(0)){
        cout << "Error abrir camaras" << endl;
        return 0;
    }

    namedWindow("Entrada"); // Ventana imagenes entrada

    int i = 0;      // Nmero de imagen
    char key = 'z'; // Key

    // Instrucciones usuario
    cout << "Pulse SPACIO para guardar las imagenes." << endl
         << "Pulse ESC para finalizar el programa." << endl << endl
         << "RECUERDE: pulse la tecla siempre que tenga seleccionada" << endl
         << "cualquiera de las ventanas Izquierda o Derecha, si " << endl
         << "intenta escribir en la ventana de comandos no funcionara." << endl;

    // Bucle de trabajo
    while(key != 27){
        // Adquisicin fotogramas
        capIzda.grab();
        capDcha.grab();

        // Asignacin fotogramas a Mats
        Mat izqFrame, dchaFrame;                    // Imagen izquierda y derecha
        capIzda.retrieve(izqFrame);
        capDcha.retrieve(dchaFrame);

        if(dchaFrame.empty() || izqFrame.empty()){
        cout << "izqFrame o dchaFrame vacia" << endl;
        return 0;
        }

        flip(dchaFrame,dchaFrame,-1);

        // Combinacion imagen izquierda + derecha
        izqFrame.copyTo(combo(Range::all(), Range(0, FRAME_WIDTH)));
        dchaFrame.copyTo(combo(Range::all(), Range(FRAME_WIDTH, 2*FRAME_WIDTH)));

        imshow("Entrada", combo);   // Visualizacion imagenes

        // Guardado de imagenes
        if(key == ' '){                         // Aumenta nmero de imagen
            i += 1;
            stringstream izq_name, dcha_name;   // Declaracin nombres imagenes
            izq_name << "Izquierda" << i << ".jpg";
            dcha_name << "Derecha" << i << ".jpg";

            // Escritura de imagenes
            imwrite(string("../Data/Imagenes/")+izq_name.str(),izqFrame);
            imwrite(string("../Data/Imagenes/")+dcha_name.str(),dchaFrame);

            izq_imagenes.push_back(izqFrame);     // Aadir imagen vector imagenes
            dcha_imagenes.push_back(dchaFrame);

            cout << "Imagenes " << i << "\tguardadas" << endl;  // Feedback usunario
        }
        key = (char)waitKey(30);    // Lectura de pulsacion teclado
    }

    destroyAllWindows();


    char visIma;
    cout << "\nQuiere visualizar imagenes guardadas (y/n): ";
    cin >> visIma;
    if(visIma =='y'){
        namedWindow("Imagenes capturadas");
        for(unsigned i = 0; i < izq_imagenes.size(); i++)
        {
            izq_imagenes[i].copyTo(combo(Range::all(), Range(0, FRAME_WIDTH)));
            dcha_imagenes[i].copyTo(combo(Range::all(), Range(FRAME_WIDTH, 2*FRAME_WIDTH)));

            imshow("Imagenes capturadas",combo);
            waitKey(0);
        }
        destroyAllWindows();
    }


    cout << "Imagenes leidas: " << izq_imagenes.size() << endl;

    float tam_lado = 18.f;  // En milimetros
    int esq_int_ancho = 9;  // Numero de esquinas ancho
    int esq_int_alto = 5;   // Numero de esquinas alto
    char opcion;

    // Posible modificacion de valores por usuario
    cout << "\nTamanio lado del cuadrado en mm por defecto = " << tam_lado << endl
         << "Numero de esquinas ancho por defecto = " << esq_int_ancho << endl
         << "Numero de esquinas alto por defecto = " << esq_int_alto << endl
         << "Desea cambiar los valores(y/n): ";
    cin >> opcion;

    if(opcion == 'y'){
        cout << "Introduzca tamanio lado en mm: ";
        cin >> tam_lado;

        cout << "Introduzca numero de esquinas ancho: ";
        cin >> esq_int_ancho;

        cout << "Introduzca numero de esquinas alto: ";
        cin >> esq_int_alto;
    }

    vector<Point3f> ob_p;
    for(int i = 0; i < esq_int_alto; i++) {
        for(int j = 0; j < esq_int_ancho; j++) {
            ob_p.push_back(Point3f(j * tam_lado, i * tam_lado, 0.f));
        }
    }

    cout << "\nDesea observar el resultado de la busqueda de esquinas(y/n): ";
    cin >> opcion;

    Mat izq_ima, dcha_ima;                      // Imagen izquierda y derecha
    vector<vector<Point2f> > izq_ima_puntos, dcha_ima_puntos;
    vector<vector<Point3f> > puntos_objetivo;
    for(unsigned i = 0; i < izq_imagenes.size(); i++) {
        // Asignacion de imagenes
        izq_ima = izq_imagenes[i];
        dcha_ima = dcha_imagenes[i];

        vector<Point2f> izq_ima_p, dcha_ima_p;  // Vectores(x,y), almacenar posiciones esquinas

        // Busqueda esquinas
        bool izq_patron_found = findChessboardCorners(izq_ima, Size(esq_int_ancho, esq_int_alto), izq_ima_p, CALIB_CB_ADAPTIVE_THRESH + CALIB_CB_NORMALIZE_IMAGE+ CALIB_CB_FAST_CHECK);
        bool dcha_patron_found = findChessboardCorners(dcha_ima, Size(esq_int_ancho, esq_int_alto), dcha_ima_p, CALIB_CB_ADAPTIVE_THRESH + CALIB_CB_NORMALIZE_IMAGE+ CALIB_CB_FAST_CHECK);

        if(izq_patron_found && dcha_patron_found) {
            puntos_objetivo.push_back(ob_p);
            Mat gray;                               // Imagen gris
            cvtColor(izq_ima, gray, CV_BGR2GRAY);   // Convertir imagen izquierda de RGB a GRIS
            cornerSubPix(gray, izq_ima_p, Size(5, 5), Size(-1, -1), TermCriteria(CV_TERMCRIT_EPS + CV_TERMCRIT_ITER, 30, 0.1));
            cvtColor(dcha_ima, gray, CV_BGR2GRAY);  // Convertir imagen derecha de RGB a GRIS
            cornerSubPix(gray, dcha_ima_p, Size(5, 5), Size(-1, -1), TermCriteria(CV_TERMCRIT_EPS + CV_TERMCRIT_ITER, 30, 0.1));
            izq_ima_puntos.push_back(izq_ima_p);    // Aadir vector puntos(x,y)
            dcha_ima_puntos.push_back(dcha_ima_p);  // Aadir vector puntos(x,y)

            if(opcion == 'y'){
                stringstream izq_corner, dcha_corner;   // Declaracin nombres imagenes
                izq_corner << "izq_corner" << i << ".jpg";
                dcha_corner << "dcha_corner" << i << ".jpg";

                Mat im_show = izq_ima.clone();                                                          // Creacin copia imagen izquierda
                drawChessboardCorners(im_show, Size(esq_int_ancho, esq_int_alto), izq_ima_p, true);     // Dibujar en imagen las esquinas
                im_show.copyTo(combo(Range::all(), Range(0, FRAME_WIDTH)));
                imwrite(string("../Data/Imagenes/")+izq_corner.str(),im_show);

                im_show = dcha_ima.clone();                                                             // Creacin copia imagen derecha
                drawChessboardCorners(im_show, Size(esq_int_ancho, esq_int_alto), dcha_ima_p, true);    // Dibujar en imagen las esquinas
                im_show.copyTo(combo(Range::all(), Range(FRAME_WIDTH, 2*FRAME_WIDTH)));
                imwrite(string("../Data/Imagenes/")+dcha_corner.str(),im_show);

                namedWindow("Combo Imagenes Leidas");   // Declaracion ventana
                imshow("Combo Imagenes Leidas",combo);  // Mostrar imagen

                cout << "Pulse cualquier tecla para continuar..." << endl;
                waitKey(0);
            }
        } else {
            izq_imagenes.erase(izq_imagenes.begin() + i);
            dcha_imagenes.erase(dcha_imagenes.begin() + i);
        }
    }
    destroyAllWindows();

    /// Rectificacion
    Mat cameraMatrix, distCoeffs;
    vector<Mat> rvecs, tvecs;
    cout << "\nError cometido en calibracion" << endl;

    float rms_error_1 = calibrateCamera(puntos_objetivo, izq_ima_puntos, izq_imagenes[0].size(), cameraMatrix, distCoeffs, rvecs, tvecs);
    cout << "RMS error 1" << rms_error_1 << endl;

    Mat R, T, E, F;
    double rms_error_2 = stereoCalibrate(puntos_objetivo,izq_ima_puntos,dcha_ima_puntos,cameraMatrix,distCoeffs,cameraMatrix,distCoeffs,izq_imagenes[0].size(),R,T,E,F);
    cout << "RMS error 2" << rms_error_2 << endl;

    Mat Rl, Rr, Pl, Pr, Q;
    Size imagen_size(FRAME_WIDTH, FRAME_HEIGHT);
    stereoRectify(cameraMatrix,distCoeffs,cameraMatrix,distCoeffs,imagen_size,R,T,Rl,Rr,Pl,Pr,Q);

    Mat map_l1, map_l2, map_r1, map_r2;
    initUndistortRectifyMap(cameraMatrix,distCoeffs, Rl,Pl,imagen_size, CV_16SC2, map_l1,map_l2);
    initUndistortRectifyMap(cameraMatrix,distCoeffs, Rr,Pr,imagen_size, CV_16SC2, map_r1,map_r2);

    cout << "\nDesea visualizar la calibracion y rectificacion(y/n): ";
    cin >> opcion;

    if(opcion == 'y'){
        destroyAllWindows();    // Destruir todas las ventanas

        VideoCapture capIzda(1), capDcha(0);    // Declaracion camaras

        // Set tamao imagenes capturadas
        capIzda.set(CV_CAP_PROP_FRAME_HEIGHT, FRAME_HEIGHT);
        capIzda.set(CV_CAP_PROP_FRAME_WIDTH, FRAME_WIDTH);
        capDcha.set(CV_CAP_PROP_FRAME_HEIGHT, FRAME_HEIGHT);
        capDcha.set(CV_CAP_PROP_FRAME_WIDTH, FRAME_WIDTH);

        namedWindow("Imagenes rectificadas");   // Nueva ventana

        // Bucle visualizacion imagenes calibradas y rectificadas
        cout << "\nPulse tecla ESC para salir de la visualizacion" << endl;
        int j = 0;
        while(char(waitKey(1)) != 27) {
            capIzda.grab();     // Captura de frame camara izquierda
            capDcha.grab();     // Captura de frame camara derecha

            Mat framel, framel_rect, framer, framer_rect;
            capIzda.retrieve(framel);   // Asignacion frame a Mat framel
            capDcha.retrieve(framer);   // Asignacion frame a Mat framer

            if(framel.empty() || framer.empty())    break;  // Comprobacion Mats son imagenes

            flip(framer,framer,-1);

            // Remapeo de las imagenes
            remap(framel, framel_rect, map_l1, map_l2, INTER_LINEAR);
            remap(framer, framer_rect, map_r1, map_r2, INTER_LINEAR);

            // Creacion imagen mayor, combinacion imagen izquierda + derecha
            Mat combo(FRAME_HEIGHT, 2 * FRAME_WIDTH, CV_8UC3);
            framel_rect.copyTo(combo(Range::all(), Range(0, FRAME_WIDTH)));
            framer_rect.copyTo(combo(Range::all(), Range(FRAME_WIDTH, 2*FRAME_WIDTH)));

            // Dibujar lineas horizontales observar rectificado
            for(int y = 0; y < combo.rows; y += 20)
                line(combo, Point(0, y), Point(combo.cols, y), Scalar(0, 0, 255));

            imshow("Imagenes rectificadas", combo); // Mostrar imagen

            stringstream comborec;   // Declaracin nombres imagenes
            comborec << "comborec" << j << ".jpg";

            imwrite(string("../Data/Imagenes/PruebaCorreccion")+comborec.str(),combo);

            j++;
        }
    }

    /// Escritura fichero
    string filecalibracion = "../Data/Calibracion.yml";   // Nombre fichero
    FileStorage fs(filecalibracion,FileStorage::WRITE);
    time_t rawtime;
    time(&rawtime);

    // Campos escritos en fichero
    fs << "Date" << asctime(localtime(&rawtime));
    fs << "cameraMatrix" << cameraMatrix;
    fs << "distCoeffs" << distCoeffs;
    fs << "R" << R;
    fs << "T" << T;
    fs << "E" << E;
    fs << "F" << F;
    fs << "Rl" << Rl;
    fs << "Rr" << Rr;
    fs << "Pl" << Pl;
    fs << "Pr" << Pr;
    fs << "Q" << Q;
    fs << "map_l1" << map_l1;
    fs << "map_l2" << map_l2;
    fs << "map_r1" << map_r1;
    fs << "map_r2" << map_r2;

    fs.release();

    return 0;
}
