#include "opencv2/core/core.hpp"
#include "opencv2/calib3d/calib3d.hpp"
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <iostream>
#include "../include/config.h"

using namespace cv;
using namespace std;

int main()
{
    /// Lectura fichero calibracion
    Mat map_r1, map_r2; //pixel maps para x e y para rectificar imagen Dcha
    Mat map_l1, map_l2; //pixel maps para x e y para rectificar imagen Dcha
    Mat Q;              //necesaria para remapeo 3D

    string fileCalib = "../Data/Calibracion.yml";   // Directorio fichero
    FileStorage fs_cal(fileCalib, FileStorage::READ);
    fs_cal["Q"] >> Q;
    fs_cal["map_l1"] >> map_l1;
    fs_cal["map_l2"] >> map_l2;
    fs_cal["map_r1"] >> map_r1;
    fs_cal["map_r2"] >> map_r2;

    fs_cal.release();

    // Comprobacion lectura valores
    if(map_r1.empty() || map_r2.empty() || map_l1.empty() || map_l2.empty() || Q.empty())
    {
        cout << "ERROR al cargar fichero de calibracion" << endl;
        return -1;
    }

    /// Configuracion camaras
    VideoCapture capIzda(1), capDcha(0);
    capIzda.set(CV_CAP_PROP_FRAME_HEIGHT, FRAME_HEIGHT);
    capIzda.set(CV_CAP_PROP_FRAME_WIDTH, FRAME_WIDTH);
    capDcha.set(CV_CAP_PROP_FRAME_HEIGHT, FRAME_HEIGHT);
    capDcha.set(CV_CAP_PROP_FRAME_WIDTH, FRAME_WIDTH);

    if(!capIzda.open(1) || !capDcha.open(0)){
        cout << "Error abrir camaras" << endl;
        return 0;
    }

    /// Inicializacion StereoSGBM
    int opcion = 0;
    string fileSGBM = "../Data/ParametrosStereo.xml";

    cout << "Seleccione una de las opciones:\n" <<
            "Leer valores de fichero -> 1\n" <<
            "Utilizar valores estandar -> 2\n" <<
            "Opcion: ";

    cin >> opcion;



    int minDisparity, numDisparity, SADWindowSize, P1, P2, disp12MaxDiff, preFilterCap, uniquenessRatio, speckleWindow, speckleRange;
    int grabar = 0;

    if(opcion == 2){
        minDisparity    = 30;
        numDisparity    = 2;
        SADWindowSize   = 3;
        P1              = 8 * 3 * SADWindowSize * SADWindowSize;
        P2              = 32 * 3 * SADWindowSize * SADWindowSize;
        disp12MaxDiff   = 1;
        preFilterCap    = 63;
        uniquenessRatio = 10;
        speckleWindow   = 100;
        speckleRange    = 32;

    }else if(opcion == 1){
        FileStorage fs_BM(fileSGBM, FileStorage::READ);

        fs_BM["minDisparity"] >> minDisparity;
        fs_BM["numDisparity"] >> numDisparity;
        fs_BM["SADWindowSize"] >> SADWindowSize;
        fs_BM["P1"] >> P1;
        fs_BM["P2"] >> P2;
        fs_BM["disp12MaxDiff"] >> disp12MaxDiff;
        fs_BM["preFilterCap"] >> preFilterCap;
        fs_BM["uniquenessRatio"] >> uniquenessRatio;
        fs_BM["speckleWindow"] >> speckleWindow;
        fs_BM["speckleRange"] >> speckleRange;
        fs_BM.release();
    }else{ return -1;}

    //Inicializacion sliders
    namedWindow("ValorStereoBM");
    resizeWindow("ValorStereoBM",FRAME_WIDTH,FRAME_HEIGHT);
    createTrackbar("minDisp","ValorStereoBM", &minDisparity, 200);
    createTrackbar("numDisp","ValorStereoBM", &numDisparity, 20);
    createTrackbar("SADWind","ValorStereoBM", &SADWindowSize, 11);
    createTrackbar("disp12MaxDiff","ValorStereoBM", &disp12MaxDiff, 100);
    createTrackbar("preFilterCap","ValorStereoBM", &preFilterCap, 200);
    createTrackbar("uniquenessRatio","ValorStereoBM", &uniquenessRatio, 50);
    createTrackbar("speckleWindows","ValorStereoBM", &speckleWindow, 300);
    createTrackbar("speckleRange","ValorStereoBM", &speckleRange, 100);
    createTrackbar("Grabar","ValorStereoBM", &grabar, 1);


    /// Bucle principal
    while(1){
        /// Toma de imagenes
        // Captura de frames
        capIzda.grab();
        capDcha.grab();

        // Asignacion de frames a imagenes
        Mat izqFrame, dchaFrame;
        capIzda.retrieve(izqFrame);
        capDcha.retrieve(dchaFrame);

        // Comprobacion imagenes obtenidas
        if(izqFrame.empty() || dchaFrame.empty()){
            cout << "ERROR. Imagenes de camaras." << endl;
            return 0;
        }

        flip(dchaFrame,dchaFrame,-1);

        /// Rectificado
        Mat frameIzdaRect, frameDchaRect;
        remap(izqFrame, frameIzdaRect, map_l1, map_l2, INTER_LINEAR);
        remap(dchaFrame, frameDchaRect, map_r1, map_r2, INTER_LINEAR);

        //Combinacion de imagenes de entrada
        Mat entrada(FRAME_HEIGHT, 2 * FRAME_WIDTH, CV_8UC3);
        frameIzdaRect.copyTo(entrada(Range::all(), Range(0, FRAME_WIDTH)));
        frameDchaRect.copyTo(entrada(Range::all(), Range(FRAME_WIDTH, 2*FRAME_WIDTH)));

        // Visualizacion entrada
        namedWindow("Entrada");
        imshow("Entrada", entrada);   // Visualizacion imagenes

        /// Correspondencia estereoscopica
        Mat disp, disp8;    // Declaracion imagenes disparidad

        // Correccion de valores stereoSGBM
        if(numDisparity < 1) numDisparity = 1;
        if(SADWindowSize < 3) SADWindowSize = 3;

        // Declaracion y computacion de stereoSGBM
        Ptr<StereoSGBM> sgbm = StereoSGBM::create(minDisparity,numDisparity*16,SADWindowSize,P1,P2,disp12MaxDiff,preFilterCap,uniquenessRatio,speckleWindow,speckleRange,true);
        sgbm->compute(frameIzdaRect,frameDchaRect, disp);
        normalize(disp, disp8, 0, 255, CV_MINMAX, CV_8U);   // Mapa de disparidad

        // Visualizacion mapa disparidad
        namedWindow("Resultado SGBM");
        imshow("Resultado SGBM",disp8);

        if(waitKey(30) == 27) {
            if(grabar == 1){
                //string fileSGBM = "../Data/ParametrosStereo.xml";
                FileStorage fs(fileSGBM,FileStorage::WRITE);
                time_t rawtime;
                time(&rawtime);
                fs << "Date" << asctime(localtime(&rawtime));
                fs << "minDisparity" << minDisparity;
                fs << "numDisparity" << numDisparity;
                fs << "SADWindowSize" << SADWindowSize;
                fs << "P1" << P1;
                fs << "P2" << P2;
                fs << "disp12MaxDiff" << disp12MaxDiff;
                fs << "preFilterCap" << preFilterCap;
                fs << "uniquenessRatio" << uniquenessRatio;
                fs << "speckleWindow" << speckleWindow;
                fs << "speckleRange" << speckleRange;

                fs.release();
                cout << "Guardado OK" << endl;
            }
            break;
        }
    }
    return(0);
}



