/// Programa: FiltroColor
/// Descripcion: Realiza la seleccion del color a filtrar en la imagen,
/// esta se puede realizar mediante sliders o cliqueando en la imagen sobre
/// el color que se quiere seleccionar.
/// Autor: Carlos Castedo Hernandez
/// Fecha: 12/03/2017

#include "opencv2/core/core.hpp"
#include "opencv2/calib3d/calib3d.hpp"
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <stdio.h>
#include <string.h>
#include <iostream>
#include <time.h>
#include "../include/config.h"

using namespace cv;
using namespace std;

/// Variables globales para selector color
int pos_x = 0,  // Coordenada x en imagen
    pos_y = 0;  // Cordenada y en imagen
bool make = false;

void CallBackFunc(int event, int x, int y, int flags, void* userdata);
void RGBtoHSV(int r, int g, int b, int &h, int &s, int &v);
void HSVtoRGB(int h, int s, int v, int &r, int &g, int &b);
void setToZero(int &ha, int &hb, int &hc, int &la, int &lb, int &lc);

int main()
{
    /// Configuracion camaras
    VideoCapture cam(0);
    cam.set(CV_CAP_PROP_FRAME_HEIGHT, FRAME_HEIGHT);
    cam.set(CV_CAP_PROP_FRAME_WIDTH, FRAME_WIDTH);

    /// Seleccion metodo de toma de color
    cout << "Metodos seleccion del color \n" <<
            "   Sliders(HSV) -> 1 \n" <<
            "   Selector color(RGB) -> 2 \n" <<
            "Elija una opcion: ";
    int opcion;
    cin >> opcion;

    Mat element = getStructuringElement( MORPH_ELLIPSE, Size(TAM_STREL,TAM_STREL));

    /// Variables del color
    int H = 0, S = 0, V = 0;

    int hH = 0, hS = 0, hV = 0,
        lH = 0, lS = 0, lV = 0;

    int R = 0, B = 0, G = 0;

    int hR = 0, hG = 0, hB = 0,
        lR = 0, lG = 0, lB = 0;


    /// Variables
    int rangeValue = 10,
        change_rangeValue = 0,
        tamDiag = 4,
        change_tamDiag = 0,
        grabar = 0;

    namedWindow("Salida");
    namedWindow("Parametros");
    namedWindow("Entrada");

    setMouseCallback("Entrada", CallBackFunc, NULL);

    /// Sliders
    if(opcion == 2){
        resizeWindow("Parametros", FRAME_WIDTH, 140);
        createTrackbar("Rango","Parametros", &rangeValue, 127);
        createTrackbar("Tamanio Muestra","Parametros", &tamDiag, 25);
        createTrackbar("Grabar","Parametros", &grabar, 1);

    }else if(opcion == 1){
        resizeWindow("Parametros", FRAME_WIDTH, 325);
        createTrackbar("High H","Parametros", &hH, 359);
        createTrackbar("Low H","Parametros", &lH, 359);
        createTrackbar("High S","Parametros", &hS, 255);
        createTrackbar("Low S","Parametros", &lS, 255);
        createTrackbar("High V","Parametros", &hV, 255);
        createTrackbar("Low V","Parametros", &lV, 255);
        createTrackbar("Grabar","Parametros", &grabar, 1);
    }

    setMouseCallback("Entrada", CallBackFunc, NULL);

    while(1){
        /// Captura de imagen y comprobacion
        Mat entrada, resultado;
        cam.retrieve(entrada);
        if(entrada.empty()) {
            cout << "Error conexion camara" << endl;
            break;
        }

        if(opcion == 1){
            /// Metodo sliders
            if(make){
                Vec3b colorHSV;
                int n_pos_x = pos_x - (tamDiag/2);
                int n_pos_y = pos_y - (tamDiag/2);

                for(int i = 0; i < tamDiag; i++){
                    for(int j = 0; j< tamDiag; j++){
                        colorHSV = entrada.at<Vec3b>(n_pos_y,n_pos_x);
                        H += colorHSV.val[0];
                        S += colorHSV.val[1];
                        V += colorHSV.val[2];
                        n_pos_x += 1;
                    }
                    n_pos_y += 1;
                }
                H /= tamDiag*tamDiag;
                S /= tamDiag*tamDiag;
                V /= tamDiag*tamDiag;

                hH = H;
                lH = H;
                hS = S;
                lS = S;
                hV = V;
                lV = V;

                // Posicionamiento en valor cliqueado
                setTrackbarPos("High H", "Parametros", hH);
                setTrackbarPos("Low H", "Parametros", lH);
                setTrackbarPos("High S", "Parametros", hS);
                setTrackbarPos("Low S", "Parametros", lS);
                setTrackbarPos("High V", "Parametros", hV);
                setTrackbarPos("Low V", "Parametros", lV);

                make = false;
            }

            Mat imgHSV;
            cvtColor(entrada,imgHSV, COLOR_BGR2HSV);

            if(lH < 180){
                Mat  imgBin_179, imgBin_0;              //Img con binariz hasta 179 y a partir de 0 resp.
                Scalar HSV_179 = Scalar(179,hS,hV);     //Hasta 179
                Scalar HSV_0   = Scalar(0,lS,lV);       //De 0 en adelante
                inRange(imgHSV, Scalar(lH, lS, lV),  HSV_179, imgBin_179);
                //En inRange quitamos 180 para que trabaje en la escala 0-180 de openCV
                inRange(imgHSV, HSV_0, Scalar(hH-180, hS, hV), imgBin_0);
                add(imgBin_179, imgBin_0, resultado);
            }
            else  //si no nos hemos metido en zona la negativa con Hmin
            {
                //En inRange quitamos 180 para que trabaje en la escala 0-180 de openCV
                inRange(imgHSV, Scalar(lH-180, lS, lV), Scalar(hH-180, hS, hV), resultado);
            }
        }
        else if(opcion == 2){
            /// Metodo selector de color
            if(make||rangeValue!=change_rangeValue||tamDiag!=change_tamDiag){
                Vec3b colorRGB;
                int n_pos_x = pos_x - (tamDiag/2);
                int n_pos_y = pos_y - (tamDiag/2);

                // Calculo valor medio de pixeles
                for(int i = 0; i < tamDiag; i++){
                    for(int j = 0; j< tamDiag; j++){
                        colorRGB = entrada.at<Vec3b>(n_pos_y,n_pos_x);
                        B += colorRGB.val[0];
                        G += colorRGB.val[1];
                        R += colorRGB.val[2];
                        n_pos_x++;
                    }
                    n_pos_y++;
                }

                B /= tamDiag*tamDiag;
                G /= tamDiag*tamDiag;
                R /= tamDiag*tamDiag;

                // Rango valores
                hR = R + rangeValue,
                hG = G + rangeValue,
                hB = B + rangeValue,
                lR = R - rangeValue,
                lG = G - rangeValue,
                lB = B - rangeValue;

                if(hR > 255){ hR = 255; }
                if(hG > 255){ hG = 255; }
                if(hB > 255){ hB = 255; }
                if(lR < 0){ lR = 0; }
                if(lG < 0){ lG = 0; }
                if(lB < 0){ lB = 0; }

                make = false; // Condicion para que solo cambie con clic
                change_rangeValue = rangeValue;
                change_tamDiag = tamDiag;
            }
            inRange(entrada,Scalar(lB,lG,lR),Scalar(hB,hG,hR),resultado);
        }else{
            cout << "Eleccion incorrecta" << endl;
            return -1;
        }

        /// Apertura y cierre
        morphologyEx(resultado,resultado, MORPH_OPEN,  element);
        morphologyEx(resultado,resultado, MORPH_CLOSE, element);

        /// Visualizacion entrada y salida
        imshow("Entrada",entrada);
        imshow("Salida",resultado);

        if(waitKey(30) == 27){
            if(grabar == 1){
                /// Configurar valores del modelo del color
                if(opcion == 1){ setToZero(hB,hG,hR,lB,lG,lR); }
                else{ setToZero(hH,hS,hV,lH,lS,lV); }
                /// Guardado de variables en fichero
                string fileColor = DATA_FOLDER + string("ColorDetecta.xml");
                FileStorage fs(fileColor,FileStorage::WRITE);
                time_t rawtime;
                time(&rawtime);

                fs << "Date" << asctime(localtime(&rawtime));
                fs << "hH" << hH;
                fs << "hS" << hS;
                fs << "hV" << hV;
                fs << "lH" << lH;
                fs << "lS" << lS;
                fs << "lV" << lV;
                fs << "hR" << hR;
                fs << "hG" << hG;
                fs << "hB" << hB;
                fs << "lR" << lR;
                fs << "lG" << lG;
                fs << "lB" << lB;

                fs.release();
                cout << "Guardado OK" << endl;
            }
            break;
        }
    }
    cam.release();
    return 0;
}

void CallBackFunc(int event, int x, int y, int flags, void* userdata)
{
    // Permite obtener las coordenadas del cursor en la imagen cuando
    // el usuario hace clic izquierdo
    if( event == EVENT_LBUTTONDOWN ){
        pos_x = x;
        pos_y = y;
        make = true;
    }
}

void RGBtoHSV(int r, int g, int b, int &h, int &s, int &v){

    Mat paletaRGB(20,20,CV_8UC3);
    Mat paletaHSV(20,20,CV_8UC3);

    for(int y=0;y<paletaRGB.rows;y++){
        for(int x=0;x<paletaRGB.cols;x++){
            Vec3b colorRGB = paletaRGB.at<Vec3b>(Point(x,y));
            colorRGB.val[0] = b;
            colorRGB.val[1] = g;
            colorRGB.val[2] = r;
            paletaRGB.at<Vec3b>(Point(x,y)) = colorRGB;
        }
    }

    cvtColor(paletaRGB,paletaHSV,CV_BGR2HSV);

    Vec3b colorHSV = paletaHSV.at<Vec3b>(Point(0,0));
    h = colorHSV.val[0];
    s = colorHSV.val[1];
    v = colorHSV.val[2];
}

void HSVtoRGB(int h, int s, int v, int &r, int &g, int &b){

    Mat paletaRGB(20,20,CV_8UC3);
    Mat paletaHSV(20,20,CV_8UC3);

    for(int y=0;y<paletaHSV.rows;y++){
        for(int x=0;x<paletaHSV.cols;x++){
            Vec3b colorHSV = paletaHSV.at<Vec3b>(Point(x,y));
            colorHSV.val[0] = h;
            colorHSV.val[1] = s;
            colorHSV.val[2] = v;
            paletaHSV.at<Vec3b>(Point(x,y)) = colorHSV;
        }
    }

    cvtColor(paletaHSV,paletaRGB,CV_HSV2BGR);

    Vec3b colorRGB = paletaRGB.at<Vec3b>(Point(0,0));
    b = colorRGB.val[0];
    g = colorRGB.val[1];
    r = colorRGB.val[2];
}

void setToZero(int &ha, int &hb, int &hc, int &la, int &lb, int &lc){
    ha = 0;
    hb = 0;
    hc = 0;
    la = 0;
    lb = 0;
    lc = 0;
}
