#include "opencv2/core/core.hpp"
#include "opencv2/calib3d/calib3d.hpp"
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <iostream>
#include "../include/config.h"

using namespace std;
using namespace cv;

void CallBackFunc(int event, int x, int y, int flags, void* userdata);

int pos_x = 0,  // Coordenada x en imagen
    pos_y = 0;  // Cordenada y en imagen
bool make = false;

int main()
{
    /// Lectura fichero calibracion
    Mat map_r1, map_r2; //pixel maps para x e y para rectificar imagen Dcha
    Mat map_l1, map_l2; //pixel maps para x e y para rectificar imagen Dcha
    Mat Q;              //necesaria para remapeo 3D
    string fileCalib = "../Data/Calibracion.yml";
    FileStorage fs_cal(fileCalib, FileStorage::READ);
    fs_cal["Q"] >> Q;
    fs_cal["map_l1"] >> map_l1;
    fs_cal["map_l2"] >> map_l2;
    fs_cal["map_r1"] >> map_r1;
    fs_cal["map_r2"] >> map_r2;

    fs_cal.release();

    // Comprobacion lectura valores
    if(map_r1.empty() || map_r2.empty() || map_l1.empty() || map_l2.empty())
    {
        cout << "ERROR al cargar fichero de calibracion" << endl;
        return -1;
    }

    /// Lectura fichero rango colores
    int hH,hS,hV,lH,lS,lV,hR,hG,hB,lR,lG,lB;
    bool modeHSV = true;
    string fileFiltro = "../Data/ColorDetecta.xml";
    FileStorage fs_fil(fileFiltro, FileStorage::READ);
    fs_fil["hH"] >> hH;
    fs_fil["hS"] >> hS;
    fs_fil["hV"] >> hV;
    fs_fil["lH"] >> lH;
    fs_fil["lS"] >> lS;
    fs_fil["lV"] >> lV;
    fs_fil["hR"] >> hR;
    fs_fil["hG"] >> hG;
    fs_fil["hB"] >> hB;
    fs_fil["lR"] >> lR;
    fs_fil["lG"] >> lG;
    fs_fil["lB"] >> lB;
    fs_fil.release();

    if(hH == 0 && hS == 0 && hV == 0 && lH == 0 && lS == 0 && lV == 0){
        modeHSV = false;
    }

    /// Lectura fichero parametros stereoSGBM
    int minDisparity, numDisparity, SADWindowSize, P1, P2, disp12MaxDiff, preFilterCap, uniquenessRatio, speckleWindow, speckleRange;

    string fileSGBM = "../Data/ParametrosStereo.xml";
    FileStorage fs_BM(fileSGBM, FileStorage::READ);

    fs_BM["minDisparity"] >> minDisparity;
    fs_BM["numDisparity"] >> numDisparity;
    fs_BM["SADWindowSize"] >> SADWindowSize;
    fs_BM["P1"] >> P1;
    fs_BM["P2"] >> P2;
    fs_BM["disp12MaxDiff"] >> disp12MaxDiff;
    fs_BM["preFilterCap"] >> preFilterCap;
    fs_BM["uniquenessRatio"] >> uniquenessRatio;
    fs_BM["speckleWindow"] >> speckleWindow;
    fs_BM["speckleRange"] >> speckleRange;
    fs_BM.release();

    ///Setup adquisicion imagen
    VideoCapture capIzda(1), capDcha(0);
    capIzda.set(CV_CAP_PROP_FRAME_HEIGHT, FRAME_HEIGHT);
    capIzda.set(CV_CAP_PROP_FRAME_WIDTH, FRAME_WIDTH);
    capDcha.set(CV_CAP_PROP_FRAME_HEIGHT, FRAME_HEIGHT);
    capDcha.set(CV_CAP_PROP_FRAME_WIDTH, FRAME_WIDTH);

    Mat element = getStructuringElement( MORPH_ELLIPSE, Size(TAM_STREL,TAM_STREL));

    namedWindow("Mapa disparidad, prefiltrado");
    namedWindow("Mapa disparidad, postfiltrado");
    setMouseCallback("Mapa disparidad, postfiltrado", CallBackFunc, NULL);
    namedWindow("Izquierda");
    namedWindow("Derecha");

    while(1){
        /// Captura de imagenes
        capIzda.grab();
        capDcha.grab();

        Mat frameIzda, frameIzdaRect, frameDcha, frameDchaRect;
        capIzda.retrieve(frameIzda);
        capDcha.retrieve(frameDcha);

        if(frameIzda.empty() || frameDcha.empty()){
            cout << "Error en camaras" << endl;
            return 0;
        }

        flip(frameDcha,frameDcha,-1);

        /// Rectificado
        remap(frameIzda, frameIzdaRect, map_l1, map_l2, INTER_LINEAR);
        remap(frameDcha, frameDchaRect, map_r1, map_r2, INTER_LINEAR);

        /// Color
        Mat izqmascara;
        if(modeHSV){
            /// Conversion de RGB a HSV
            Mat modelColor_izda;
            cvtColor(frameIzdaRect,modelColor_izda,COLOR_BGR2HSV);  //Pasa izda a HSV

            /// Distincion colores
            inRange(modelColor_izda,Scalar(lH,lS,lV),Scalar(hH,hS,hV),izqmascara);

        }else{
            /// Distincion colores
            inRange(frameIzdaRect,Scalar(lB,lG,lR),Scalar(hB,hG,hR),izqmascara);
        }

        /// Apertura imagenes resultado
        morphologyEx(izqmascara,izqmascara, MORPH_OPEN, element);
        morphologyEx(izqmascara,izqmascara, MORPH_CLOSE, element);

        /// StereoSGBM
        Mat disp, disp8, disp_compute;    // Declaracion imagenes StereoSGBM

        // Declaracion y computacion de stereoSGBM
        Ptr<StereoSGBM> sgbm = StereoSGBM::create(minDisparity,numDisparity*16,SADWindowSize,P1,P2,disp12MaxDiff,preFilterCap,uniquenessRatio,speckleWindow,speckleRange,true);
        sgbm->compute(frameIzdaRect, frameDchaRect, disp);
        normalize(disp, disp8, 0, 255, CV_MINMAX, CV_8U);
        disp.convertTo(disp_compute,CV_32F, 1.f/16.f);

        // Visualizacion StereoSGBM
        Mat mapaDispFil;
        disp8.copyTo(mapaDispFil,izqmascara);
        imshow("Mapa disparidad, postfiltrado",mapaDispFil);

        /// Imagen 3D
        Mat ima3D;
        reprojectImageTo3D(disp_compute,ima3D,Q,true);

        if(make == true){
            ima3D = ima3D(Range(pos_y-5,pos_y+5),Range(pos_x-5,pos_x+5));
            Mat z_roi(ima3D.size(), CV_32FC1);
            int from_to[] = {2,0};
            mixChannels(&ima3D, 1, &z_roi, 1, from_to, 1);
            cout << "Distancia: " << mean(z_roi) << " mm" << endl;

            make = false;
        }

        if(waitKey(30) == 27){ break; }
    }
    capDcha.release();
    capIzda.release();
    return 0;
}

void CallBackFunc(int event, int x, int y, int flags, void* userdata)
{
    // Permite obtener las coordenadas del cursor en la imagen cuando
    // el usuario hace clic izquierdo
    if( event == EVENT_LBUTTONDOWN ){
        pos_x = x;
        pos_y = y;
        make = true;
    }
}
