#include "opencv2/core/core.hpp"
#include "opencv2/calib3d/calib3d.hpp"
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <iostream>
#include "../include/config.h"

using namespace std;
using namespace cv;

int main()
{
    /// Lectura fichero calibracion
    Mat map_r1, map_r2; //pixel maps para x e y para rectificar imagen Dcha
    Mat map_l1, map_l2; //pixel maps para x e y para rectificar imagen Dcha
    Mat Q;              //necesaria para remapeo 3D

    string fileCalib = "../Data/Calibracion.yml";
    FileStorage fs_cal(fileCalib, FileStorage::READ);
    fs_cal["Q"] >> Q;
    fs_cal["map_l1"] >> map_l1;
    fs_cal["map_l2"] >> map_l2;
    fs_cal["map_r1"] >> map_r1;
    fs_cal["map_r2"] >> map_r2;

    fs_cal.release();

    // Comprobacion lectura valores
    if(map_r1.empty() || map_r2.empty() || map_l1.empty() || map_l2.empty())
    {
        cout << "ERROR al cargar fichero de calibracion" << endl;
        return -1;
    }

    ///Setup adquisicion imagen
    VideoCapture capIzda(0), capDcha(1);
    capIzda.set(CV_CAP_PROP_FRAME_HEIGHT, FRAME_HEIGHT);
    capIzda.set(CV_CAP_PROP_FRAME_WIDTH, FRAME_WIDTH);
    capDcha.set(CV_CAP_PROP_FRAME_HEIGHT, FRAME_HEIGHT);
    capDcha.set(CV_CAP_PROP_FRAME_WIDTH, FRAME_WIDTH);

    Mat element = getStructuringElement( MORPH_ELLIPSE, Size(TAM_STREL,TAM_STREL));

    float i = 2.0;
    int it = 0;
    float media = 0;

    int numDisparity = 0,
        blockSize = 25;

    namedWindow("Mapa Disparidad",WINDOW_AUTOSIZE);
    namedWindow("Parametros");
    resizeWindow("Parametros",FRAME_WIDTH,80);
    createTrackbar("numDisp","Parametros", &numDisparity, 80);
    createTrackbar("blockSize","Parametros", &blockSize, 50);
    cout << "Calculando..." << endl;

    const clock_t begin_time = clock();

     while(1){
        /// Captura de imagenes
        capIzda.grab();
        capDcha.grab();

        Mat frameIzda, frameIzdaRect, frameDcha, frameDchaRect;
        capIzda.retrieve(frameIzda);
        capDcha.retrieve(frameDcha);

        if(frameIzda.empty() || frameDcha.empty()){
            cout << "Error en camaras" << endl;
            return 0;
        }

        flip(frameDcha,frameDcha,-1);

        /// Rectificado
        remap(frameIzda, frameIzdaRect, map_l1, map_l2, INTER_LINEAR);
        remap(frameDcha, frameDchaRect, map_r1, map_r2, INTER_LINEAR);

        cvtColor(frameDchaRect, frameDchaRect,COLOR_BGR2GRAY);
        cvtColor(frameIzdaRect, frameIzdaRect,COLOR_BGR2GRAY);

        /// StereoBM
        Mat disp, disp8, disp_compute;    // Declaracion imagenes StereoSGBM

        // Declaracion y computacion de stereoSGBM
        if(blockSize%2 == 0){ blockSize += 1; }
        Ptr<StereoBM> bm = StereoBM::create(numDisparity*16,blockSize);
        bm->compute(frameIzdaRect, frameDchaRect, disp);
        normalize(disp, disp8, 0, 255, CV_MINMAX, CV_8U);

        // Visualizacion mapa disparidad
        imshow("Mapa Disparidad",disp8);

        it++;

        /// Calculo FPS
        float tiempo = float( clock () - begin_time ) / CLOCKS_PER_SEC;
        if(tiempo > i){
             media += it;
             it = 0;
             i = i + 1.0;
            if(i == 60.0){
                cout << "\n  El numero de frames medios por segundo obtenidos durante   \n" <<
                        "  un tiempo de ejecucion de 60 segundos es: " << media/i << endl;
                break;
            }
        }

        /// Guardado imagenes
        if(waitKey(30) == 27){
            imwrite("./Entrada.png",frameDcha);
            imwrite("./Resultado.png",disp8);
            break;
        }
    }
    capDcha.release();
    capIzda.release();
    return 0;
}
