#include "opencv2/core/core.hpp"
#include "opencv2/calib3d/calib3d.hpp"
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <iostream>
#include "../include/config.h"

using namespace std;
using namespace cv;


int main()
{
    /// Lectura fichero calibracion
    Mat map_r1, map_r2; //pixel maps para x e y para rectificar imagen Dcha
    Mat map_l1, map_l2; //pixel maps para x e y para rectificar imagen Dcha
    Mat Q;              //necesaria para remapeo 3D
    string fileCalib = "../Data/Calibracion.yml";
    FileStorage fs_cal(fileCalib, FileStorage::READ);
    fs_cal["Q"] >> Q;
    fs_cal["map_l1"] >> map_l1;
    fs_cal["map_l2"] >> map_l2;
    fs_cal["map_r1"] >> map_r1;
    fs_cal["map_r2"] >> map_r2;

    fs_cal.release();

    // Comprobacion lectura valores
    if(map_r1.empty() || map_r2.empty() || map_l1.empty() || map_l2.empty())
    {
        cout << "ERROR al cargar fichero de calibracion" << endl;
        return -1;
    }

        /// Lectura fichero parametros stereoSGBM
    int minDisparity, numDisparity, SADWindowSize, P1, P2, disp12MaxDiff, preFilterCap, uniquenessRatio, speckleWindow, speckleRange;

    string fileSGBM = "../Data/ParametrosStereo.xml";
    FileStorage fs_BM(fileSGBM, FileStorage::READ);

    fs_BM["minDisparity"] >> minDisparity;
    fs_BM["numDisparity"] >> numDisparity;
    fs_BM["SADWindowSize"] >> SADWindowSize;
    fs_BM["P1"] >> P1;
    fs_BM["P2"] >> P2;
    fs_BM["disp12MaxDiff"] >> disp12MaxDiff;
    fs_BM["preFilterCap"] >> preFilterCap;
    fs_BM["uniquenessRatio"] >> uniquenessRatio;
    fs_BM["speckleWindow"] >> speckleWindow;
    fs_BM["speckleRange"] >> speckleRange;
    fs_BM.release();

    ///Setup adquisicion imagen
    VideoCapture capIzda(0), capDcha(1);
    capIzda.set(CV_CAP_PROP_FRAME_HEIGHT, FRAME_HEIGHT);
    capIzda.set(CV_CAP_PROP_FRAME_WIDTH, FRAME_WIDTH);
    capDcha.set(CV_CAP_PROP_FRAME_HEIGHT, FRAME_HEIGHT);
    capDcha.set(CV_CAP_PROP_FRAME_WIDTH, FRAME_WIDTH);

    Mat element = getStructuringElement( MORPH_ELLIPSE, Size(TAM_STREL,TAM_STREL));

    float i = 2.0;
    int it = 0;
    float media = 0;

    namedWindow("Mapa disparidad",WINDOW_AUTOSIZE);
    cout << "Calculando..." << endl;

    const clock_t begin_time = clock();

     while(1){
        /// Captura de imagenes
        capIzda.grab();
        capDcha.grab();

        Mat frameIzda, frameIzdaRect, frameDcha, frameDchaRect;
        capIzda.retrieve(frameIzda);
        capDcha.retrieve(frameDcha);

        if(frameIzda.empty() || frameDcha.empty()){
            cout << "Error en camaras" << endl;
            return 0;
        }

        flip(frameDcha,frameDcha,-1);

        /// Rectificado
        remap(frameIzda, frameIzdaRect, map_l1, map_l2, INTER_LINEAR);
        remap(frameDcha, frameDchaRect, map_r1, map_r2, INTER_LINEAR);

        /// StereoSGBM
        Mat disp, disp8;    // Declaracion imagenes StereoSGBM

        // Declaracion y computacion de stereoSGBM
        Ptr<StereoSGBM> sgbm = StereoSGBM::create(minDisparity,numDisparity*16,SADWindowSize,P1,P2,disp12MaxDiff,preFilterCap,uniquenessRatio,speckleWindow,speckleRange,true);
        sgbm->compute(frameIzdaRect, frameDchaRect, disp);
        normalize(disp, disp8, 0, 255, CV_MINMAX, CV_8U);   // Mapa disparidad

        // Visualizacion Mapa disparidad
        imshow("Mapa disparidad",disp8);

        it++;

        /// Calculo FPS
        float tiempo = float( clock () - begin_time ) / CLOCKS_PER_SEC;
        if(tiempo > i){
             media += it;
             it = 0;
             i = i + 1.0;
            if(i == 60.0){
                cout << "\n  El numero de frames medios por segundo obtenidos durante   \n" <<
                        "  un tiempo de ejecucion de 60 segundos es: " << media/i << endl;
                break;
            }
        }

        /// Guardado de imagenes
        if(waitKey(30) == 27){
            imwrite("./Entrada.png",frameDcha);
            imwrite("./Resultado.png",disp8);
            break;
        }
    }
    capDcha.release();
    capIzda.release();
    return 0;
}

